from __future__ import annotations

from diracx.core.config.schema import Config
from diracx.core.models import GroupInfo, Metadata, OpenIDConfiguration
from diracx.core.settings import AuthSettings


async def get_openid_configuration(
    token_endpoint: str,
    userinfo_endpoint: str,
    authorization_endpoint: str,
    device_authorization_endpoint: str,
    revoke_refresh_token_endpoint: str,
    jwks_endpoint: str,
    config: Config,
    settings: AuthSettings,
) -> OpenIDConfiguration:
    """OpenID Connect discovery endpoint."""
    scopes_supported = []
    for vo in config.Registry:
        scopes_supported.append(f"vo:{vo}")
        scopes_supported += [f"group:{vo}" for vo in config.Registry[vo].Groups]
    scopes_supported += [f"property:{p}" for p in settings.available_properties]

    return {
        "issuer": settings.token_issuer,
        "token_endpoint": token_endpoint,
        "userinfo_endpoint": userinfo_endpoint,
        "authorization_endpoint": authorization_endpoint,
        "device_authorization_endpoint": device_authorization_endpoint,
        "revocation_endpoint": revoke_refresh_token_endpoint,
        "jwks_uri": jwks_endpoint,
        "grant_types_supported": [
            "authorization_code",
            "urn:ietf:params:oauth:grant-type:device_code",
        ],
        "scopes_supported": scopes_supported,
        "response_types_supported": ["code"],
        "token_endpoint_auth_signing_alg_values_supported": settings.token_allowed_algorithms,
        "token_endpoint_auth_methods_supported": ["none"],
        "code_challenge_methods_supported": ["S256"],
    }


async def get_jwks(settings: AuthSettings) -> dict:
    """Get the JWKs (public keys)."""
    return settings.token_keystore.jwks.as_dict(  # type: ignore[return-value]
        private=False,
    )


async def get_installation_metadata(
    config: Config,
) -> Metadata:
    """Get metadata about the dirac installation."""
    metadata: Metadata = {
        "virtual_organizations": {},
    }
    for vo, vo_info in config.Registry.items():
        groups: dict[str, GroupInfo] = {
            group: {"properties": sorted(group_info.Properties)}
            for group, group_info in vo_info.Groups.items()
        }
        metadata["virtual_organizations"][vo] = {
            "groups": groups,
            "support": {
                "message": vo_info.Support.Message,
                "webpage": vo_info.Support.Webpage,
                "email": vo_info.Support.Email,
            },
            "default_group": vo_info.DefaultGroup,
        }

    return metadata



from __future__ import annotations

from collections import defaultdict
from typing import Any

from diracx.core.config.schema import Config
from diracx.core.properties import JOB_SHARING
from diracx.db.sql.task_queue.db import TaskQueueDB

TQ_MIN_SHARE = 0.001
PRIORITY_IGNORED_FIELDS = ("Sites", "BannedSites")


async def recalculate_tq_shares_for_entity(
    owner: str,
    owner_group: str,
    vo: str,
    config: Config,
    task_queue_db: TaskQueueDB,
):
    """Recalculate the shares for a user/userGroup combo."""
    group_properties = config.Registry[vo].Groups[owner_group].Properties
    job_share = config.Registry[vo].Groups[owner_group].JobShare
    allow_background_tqs = config.Registry[vo].Groups[owner_group].AllowBackgroundTQs
    if JOB_SHARING in group_properties:
        # If group has JobSharing just set prio for that entry, user is irrelevant
        await set_priorities_for_entity(
            owner=None,
            owner_group=owner_group,
            job_share=job_share,
            allow_background_tqs=allow_background_tqs,
            task_queue_db=task_queue_db,
        )
        return

    # Get all owners from the owner group
    owners = await task_queue_db.get_task_queue_owners_by_group(owner_group)
    num_owners = len(owners)
    # If there are no owners do now
    if num_owners == 0:
        return

    # Split the share amongst the number of owners
    entities_shares = {owner: job_share / num_owners for owner, _ in owners.items()}

    # TODO: implement the following
    # If corrector is enabled let it work it's magic
    # if enable_shares_correction:
    #     entities_shares = await self.__shares_corrector.correct_shares(
    #         entitiesShares, group=group
    #     )

    # If the user is already known and has more than 1 tq, the rest of the users don't need to be modified
    # (The number of owners didn't change)
    if owner in owners and owners[owner] > 1:
        await set_priorities_for_entity(
            owner=owner,
            owner_group=owner_group,
            job_share=entities_shares[owner],
            allow_background_tqs=allow_background_tqs,
            task_queue_db=task_queue_db,
        )
        return

    # Oops the number of owners may have changed so we recalculate the prio for all owners in the group
    for owner in owners:
        await set_priorities_for_entity(
            owner=owner,
            owner_group=owner_group,
            job_share=entities_shares[owner],
            allow_background_tqs=allow_background_tqs,
            task_queue_db=task_queue_db,
        )


async def set_priorities_for_entity(
    owner_group: str,
    job_share: float,
    allow_background_tqs: bool,
    task_queue_db: TaskQueueDB,
    owner: str | None = None,
):
    """Set the priority for a user/userGroup combo given a split share."""
    tq_dict = await task_queue_db.get_task_queue_priorities(owner_group, owner)
    if not tq_dict:
        return

    rows = await task_queue_db.retrieve_task_queues(list(tq_dict))
    prio_dict = await calculate_priority(tq_dict, rows, job_share, allow_background_tqs)
    for prio, tqs in prio_dict.items():
        await task_queue_db.set_priorities_for_entity(tqs, prio)


async def calculate_priority(
    tq_dict: dict[int, float],
    all_tqs_data: dict[int, dict[str, Any]],
    share: float,
    allow_bg_tqs: bool,
) -> dict[float, list[int]]:
    """Calculate the priority for each TQ given a share.

    :param tq_dict: dict of {tq_id: prio}
    :param all_tqs_data: dict of {tq_id: {tq_data}}, where tq_data is a dict of {field: value}
    :param share: share to be distributed among TQs
    :param allow_bg_tqs: allow background TQs to be used
    :return: dict of {priority: [tq_ids]}
    """

    def is_background(tq_priority: float, allow_bg_tqs: bool) -> bool:
        """A TQ is background if its priority is below a threshold and background TQs are allowed."""
        return tq_priority <= 0.1 and allow_bg_tqs

    # Calculate Sum of priorities of non background TQs
    total_prio = sum(
        [prio for prio in tq_dict.values() if not is_background(prio, allow_bg_tqs)]
    )

    # Update prio for each TQ
    for tq_id, tq_priority in tq_dict.items():
        if is_background(tq_priority, allow_bg_tqs):
            prio = TQ_MIN_SHARE
        else:
            prio = max((share / total_prio) * tq_priority, TQ_MIN_SHARE)
        tq_dict[tq_id] = prio

    # Generate groups of TQs that will have the same prio=sum(prios) maomenos
    tq_groups: dict[str, list[int]] = defaultdict(list)
    for tq_id, tq_data in all_tqs_data.items():
        for field in ("Jobs", "Priority") + PRIORITY_IGNORED_FIELDS:
            if field in tq_data:
                tq_data.pop(field)
        tq_hash = []
        for f in sorted(tq_data):
            tq_hash.append(f"{f}:{tq_data[f]}")
        tq_hash = "|".join(tq_hash)
        # if tq_hash not in tq_groups:
        #     tq_groups[tq_hash] = []
        tq_groups[tq_hash].append(tq_id)

    # Do the grouping
    for tq_group in tq_groups.values():
        total_prio = sum(tq_dict[tq_id] for tq_id in tq_group)
        for tq_id in tq_group:
            tq_dict[tq_id] = total_prio

    # Group by priorities
    result: dict[float, list[int]] = defaultdict(list)
    for tq_id, tq_priority in tq_dict.items():
        result[tq_priority].append(tq_id)

    return result



"""JWKS key management scripts.

See https://datatracker.ietf.org/doc/html/rfc7517 for further details.
"""

from __future__ import annotations

import argparse
import asyncio
import json
import logging
from pathlib import Path

from joserfc.jwk import JWKRegistry, Key, KeySet
from uuid_utils import uuid7

logger = logging.getLogger(__name__)

# ---------- Helpers ----------------------------------------------------------


def load_jwks(path: Path) -> KeySet:
    """Return a (possibly empty) JWKSet."""
    if path.exists():
        return KeySet.import_key_set(json.loads(path.read_text()))
    logger.warning("JWKS file %s not found – creating a new one", path)
    return KeySet(keys=[])


def save_jwks(path: Path, jwks: KeySet) -> None:
    """Write JWKSet to disk *including* private parts."""
    path.write_text(json.dumps(jwks.as_dict(private=True), indent=2))
    logger.info("JWKS written to %s", path)


def new_key(
    kty: str = "OKP",
    crv_or_size: str | int = "Ed25519",
) -> Key:
    """Create a fresh private signing key."""
    parameters = {
        "key_ops": ["sign", "verify"],
        "alg": "EdDSA",
        "kid": uuid7().hex,
    }
    return JWKRegistry.generate_key(
        key_type=kty,
        crv_or_size=crv_or_size,
        private=True,
        parameters=parameters,  # type: ignore[arg-type]
    )


# ---------- CLI --------------------------------------------------------------


async def rotate_jwk(args):
    """Rotate keys in a JWKS file by inserting a new key at index 0 (active)."""
    logger.info("Rotating JWKs...")

    crv_or_size = args.crv_or_size
    if isinstance(crv_or_size, str) and crv_or_size.isdigit():
        crv_or_size = int(crv_or_size)

    jwks_path = Path(args.jwks_path)
    jwks = load_jwks(jwks_path)

    # Current key (at index 0) is set to "verify" only
    if len(jwks.keys) > 0:
        active_key = jwks.keys[0]
        active_key_dict = active_key.as_dict(private=True)
        active_key_dict["key_ops"] = sorted(
            set(active_key_dict.get("key_ops", [])) - {"sign"}
        )
        jwks.keys[0] = JWKRegistry.import_key(active_key_dict)

    jwk = new_key(args.kty, crv_or_size)
    jwks.keys.insert(0, jwk)

    save_jwks(jwks_path, jwks)


async def delete_jwk(args):
    """Delete a JWK from a JWKS file."""
    logger.info("Deleting JWK...")

    path = Path(args.jwks_path)
    jwks = load_jwks(path)
    jwks.keys = [k for k in jwks.keys if k.get("kid") != args.kid]
    save_jwks(path, jwks)


def parse_args():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="command", required=True)

    rotate_jwk_parser = subparsers.add_parser(
        "rotate-jwk", help="Rotate JWK keys in a JWKS file"
    )
    rotate_jwk_parser.add_argument(
        "--jwks-path", required=True, help="Path to the existing (old) JWKS JSON file."
    )

    rotate_jwk_parser.add_argument(
        "--kty", default="OKP", help="Key type for the new key."
    )
    rotate_jwk_parser.add_argument(
        "--crv-or-size", default="Ed25519", help="Curve or size for the new key."
    )
    rotate_jwk_parser.set_defaults(func=rotate_jwk)

    delete_jwk_parser = subparsers.add_parser(
        "delete-jwk", help="Delete a JWK key from a JWKS file"
    )
    delete_jwk_parser.add_argument(
        "--jwks-path", required=True, help="Path to the JWKS JSON file."
    )
    delete_jwk_parser.add_argument(
        "--kid", required=True, help="Key ID (kid) of the key to delete."
    )
    delete_jwk_parser.set_defaults(func=delete_jwk)

    args = parser.parse_args()
    logger.setLevel(logging.INFO)
    asyncio.run(args.func(args))


if __name__ == "__main__":
    parse_args()



from __future__ import annotations

import hashlib
import secrets
from datetime import timedelta
from io import BytesIO
from typing import AsyncGenerator, Generator

import botocore.exceptions
import freezegun
import httpx
import pytest
import sqlalchemy

from diracx.core.exceptions import SandboxNotFoundError
from diracx.core.models import ChecksumAlgorithm, SandboxFormat, SandboxInfo, UserInfo
from diracx.core.settings import SandboxStoreSettings
from diracx.db.sql.sandbox_metadata.db import SandboxMetadataDB
from diracx.logic.jobs.sandboxes import (
    clean_sandboxes,
    get_sandbox_file,
    initiate_sandbox_upload,
)
from diracx.testing.time import mock_sqlite_time

FAKE_USER_INFO = UserInfo(
    sub="fakevo:97ae90d3-36aa-4271-becf-e61173d93fe3",
    preferred_username="fakeuser",
    dirac_group="fake_group",
    vo="fakevo",
)


@pytest.fixture
async def sandbox_metadata_db() -> AsyncGenerator[SandboxMetadataDB, None]:
    """Create a fake sandbox metadata database."""
    db = SandboxMetadataDB(db_url="sqlite+aiosqlite:///:memory:")
    async with db.engine_context():
        sqlalchemy.event.listen(db.engine.sync_engine, "connect", mock_sqlite_time)

        async with db.engine.begin() as conn:
            await conn.run_sync(db.metadata.create_all)

        yield db


@pytest.fixture
async def sandbox_settings(
    test_sandbox_settings,
) -> AsyncGenerator[SandboxStoreSettings, None]:
    """Create a fake sandbox settings."""
    async with test_sandbox_settings.lifetime_function():
        yield test_sandbox_settings


@pytest.fixture()
def frozen_time() -> Generator[freezegun.FreezeGun, None]:
    with freezegun.freeze_time("2012-01-14") as ft:
        yield ft


async def test_upload_and_clean(
    sandbox_metadata_db: SandboxMetadataDB,
    sandbox_settings: SandboxStoreSettings,
    frozen_time: freezegun.FreezeGun,
) -> None:
    """Test the full upload, download, and cleanup workflow.

    This test will create a sandbox, upload it, download it, and then
    clean it up. It will also check that the sandbox is removed from
    the S3 bucket after the cleanup.
    """
    data = secrets.token_bytes(256)
    data_digest = hashlib.sha256(data).hexdigest()
    key = f"fakevo/fake_group/fakeuser/sha256:{data_digest}.tar.zst"
    expected_pfn = f"SB:SandboxSE|/S3/sandboxes/{key}"

    sandbox_info = SandboxInfo(
        checksum_algorithm=ChecksumAlgorithm.SHA256,
        checksum=data_digest,
        size=len(data),
        format=SandboxFormat.TAR_ZST,
    )

    # Test with a new sandbox
    async with sandbox_metadata_db:
        response = await initiate_sandbox_upload(
            FAKE_USER_INFO, sandbox_info, sandbox_metadata_db, sandbox_settings
        )
    assert response.pfn == expected_pfn
    assert response.url is not None

    # Do the actual upload
    files = {"file": ("file", BytesIO(data))}
    async with httpx.AsyncClient() as httpx_client:
        response = await httpx_client.post(
            response.url, data=response.fields, files=files
        )
        response.raise_for_status()

    # Test with the same sandbox again
    async with sandbox_metadata_db:
        response = await initiate_sandbox_upload(
            FAKE_USER_INFO, sandbox_info, sandbox_metadata_db, sandbox_settings
        )
    assert response.pfn == expected_pfn
    assert response.url is None

    # Try to download the sandbox
    async with sandbox_metadata_db:
        download_response = await get_sandbox_file(
            expected_pfn, sandbox_metadata_db, sandbox_settings
        )
    async with httpx.AsyncClient() as httpx_client:
        response = await httpx_client.get(download_response.url)
        response.raise_for_status()
        assert response.content == data

    # There should be no sandboxes to remove
    async with sandbox_metadata_db:
        await clean_sandboxes(sandbox_metadata_db, sandbox_settings)

    # Try to download the sandbox
    async with sandbox_metadata_db:
        download_response = await get_sandbox_file(
            expected_pfn, sandbox_metadata_db, sandbox_settings
        )
    async with httpx.AsyncClient() as httpx_client:
        response = await httpx_client.get(download_response.url)
        response.raise_for_status()
        assert response.content == data

    # Move forward a few weeks
    frozen_time.tick(delta=timedelta(weeks=3))

    # Check that the sandbox exists in the S3 bucket
    await sandbox_settings.s3_client.head_object(
        Bucket=sandbox_settings.bucket_name, Key=key
    )

    # Now the sandbox should be removed
    async with sandbox_metadata_db:
        await clean_sandboxes(sandbox_metadata_db, sandbox_settings)

    # Check that the sandbox was actually removed from the bucket
    with pytest.raises(botocore.exceptions.ClientError, match="Not Found"):
        await sandbox_settings.s3_client.head_object(
            Bucket=sandbox_settings.bucket_name, Key=key
        )

    # Check that the sandbox was removed
    async with sandbox_metadata_db:
        with pytest.raises(SandboxNotFoundError):
            await get_sandbox_file(expected_pfn, sandbox_metadata_db, sandbox_settings)



from __future__ import annotations

import asyncio
import logging
from copy import deepcopy
from datetime import datetime, timezone

from DIRACCommon.Core.Utilities.ClassAd.ClassAdLight import ClassAd
from DIRACCommon.Core.Utilities.ReturnValues import returnValueOrRaise
from DIRACCommon.WorkloadManagementSystem.DB.JobDBUtils import (
    checkAndAddOwner,
    compressJDL,
    createJDLWithInitialStatus,
)
from DIRACCommon.WorkloadManagementSystem.Utilities.ParametricJob import (
    generateParametricJobs,
    getParameterVectorLength,
)
from pydantic import BaseModel

from diracx.core.config import Config
from diracx.core.models import (
    InsertedJob,
    JobLoggingRecord,
    JobStatus,
    UserInfo,
)
from diracx.db.sql.job.db import JobDB
from diracx.db.sql.job_logging.db import JobLoggingDB
from diracx.logic.jobs.utils import check_and_prepare_job, make_job_manifest_config

logger = logging.getLogger(__name__)


class JobSubmissionSpec(BaseModel):
    jdl: str
    owner: str
    owner_group: str
    initial_status: str
    initial_minor_status: str
    vo: str


MAX_PARAMETRIC_JOBS = 20


async def submit_jdl_jobs(
    job_definitions: list[str],
    job_db: JobDB,
    job_logging_db: JobLoggingDB,
    user_info: UserInfo,
    config: Config,
) -> list[InsertedJob]:
    """Submit a list of JDLs to the JobDB."""
    # TODO: that needs to go in the legacy adapter (Does it ? Because bulk submission is not supported there)
    for i in range(len(job_definitions)):
        job_definition = job_definitions[i].strip()
        if not (job_definition.startswith("[") and job_definition.endswith("]")):
            job_definition = f"[{job_definition}]"
        job_definitions[i] = job_definition

    if len(job_definitions) == 1:
        # Check if the job is a parametric one
        job_class_ad = ClassAd(job_definitions[0])
        result = getParameterVectorLength(job_class_ad)
        if not result["OK"]:
            # FIXME dont do this
            print("Issue with getParameterVectorLength", result["Message"])
            return result
        n_jobs = result["Value"]
        parametric_job = False
        if n_jobs is not None and n_jobs > 0:
            # if we are here, then jobDesc was the description of a parametric job. So we start unpacking
            parametric_job = True
            result = generateParametricJobs(job_class_ad)
            if not result["OK"]:
                # FIXME why?
                return result
            job_desc_list = result["Value"]
        else:
            # if we are here, then jobDesc was the description of a single job.
            job_desc_list = job_definitions
    else:
        # if we are here, then jobDesc is a list of JDLs
        # we need to check that none of them is a parametric
        for job_definition in job_definitions:
            res = getParameterVectorLength(ClassAd(job_definition))
            if not res["OK"]:
                raise ValueError(res["Message"])

            if res["Value"]:
                raise ValueError("You cannot submit parametric jobs in a bulk fashion")

        job_desc_list = job_definitions
        # parametric_job = True
        parametric_job = False

    # TODO: make the max number of jobs configurable in the CS
    if len(job_desc_list) > MAX_PARAMETRIC_JOBS:
        raise ValueError(
            f"Normal user cannot submit more than {MAX_PARAMETRIC_JOBS} jobs at once"
        )

    result = []

    if parametric_job:
        initial_status = JobStatus.SUBMITTING
        initial_minor_status = "Bulk transaction confirmation"
    else:
        initial_status = JobStatus.RECEIVED
        initial_minor_status = "Job accepted"

    try:
        submitted_job_ids = await create_jdl_jobs(
            [
                JobSubmissionSpec(
                    jdl=jdl,
                    owner=user_info.preferred_username,
                    owner_group=user_info.dirac_group,
                    initial_status=initial_status,
                    initial_minor_status=initial_minor_status,
                    vo=user_info.vo,
                )
                for jdl in job_desc_list
            ],
            job_db=job_db,
            config=config,
        )
    except ExceptionGroup as e:
        logging.exception("JDL syntax error occurred during job submission")
        raise ValueError("JDL syntax error") from e

    logging.debug(
        f'Jobs added to the JobDB", "{submitted_job_ids} for {user_info.preferred_username}/{user_info.dirac_group}'
    )

    job_created_time = datetime.now(timezone.utc)
    await job_logging_db.insert_records(
        [
            JobLoggingRecord(
                job_id=int(job_id),
                status=initial_status,
                minor_status=initial_minor_status,
                application_status="Unknown",
                date=job_created_time,
                source="JobManager",
            )
            for job_id in submitted_job_ids
        ]
    )

    # if not parametric_job:
    #     self.__sendJobsToOptimizationMind(submitted_job_ids)

    return [
        InsertedJob(
            JobID=job_id,
            Status=initial_status,
            MinorStatus=initial_minor_status,
            TimeStamp=job_created_time,
        )
        for job_id in submitted_job_ids
    ]


async def create_jdl_jobs(jobs: list[JobSubmissionSpec], job_db: JobDB, config: Config):
    """Create jobs from JDLs and insert them into the DB."""
    jobs_to_insert = {}
    jdls_to_update = {}
    inputdata_to_insert = {}
    original_jdls = []

    # generate the jobIDs first
    # TODO: should ForgivingTaskGroup be used?
    async with asyncio.TaskGroup() as tg:
        for job in jobs:
            original_jdl = deepcopy(job.jdl)
            job_manifest = returnValueOrRaise(
                checkAndAddOwner(
                    original_jdl,
                    job.owner,
                    job.owner_group,
                    job_manifest_config=make_job_manifest_config(config, job.vo),
                )
            )

            # Fix possible lack of brackets
            if original_jdl.strip()[0] != "[":
                original_jdl = f"[{original_jdl}]"

            original_jdls.append(
                (
                    original_jdl,
                    job_manifest,
                    tg.create_task(job_db.create_job(compressJDL(original_jdl))),
                )
            )

    async with asyncio.TaskGroup() as tg:
        for job, (original_jdl, job_manifest_, job_id_task) in zip(jobs, original_jdls):
            job_id = job_id_task.result()
            job_attrs = {
                "JobID": job_id,
                "LastUpdateTime": datetime.now(tz=timezone.utc),
                "SubmissionTime": datetime.now(tz=timezone.utc),
                "Owner": job.owner,
                "OwnerGroup": job.owner_group,
                "VO": job.vo,
            }

            job_manifest_.setOption("JobID", job_id)

            # 2.- Check JDL and Prepare DIRAC JDL
            job_jdl = job_manifest_.dumpAsJDL()

            # Replace the JobID placeholder if any
            if job_jdl.find("%j") != -1:
                job_jdl = job_jdl.replace("%j", str(job_id))

            class_ad_job = ClassAd(job_jdl)

            class_ad_req = ClassAd("[]")
            if not class_ad_job.isOK():
                # Rollback the entire transaction
                logging.exception(f"Error in JDL syntax for job JDL: {original_jdl}")
                raise ValueError(f"Error in JDL syntax for job JDL: {original_jdl}")
            # TODO: check if that is actually true
            if class_ad_job.lookupAttribute("Parameters"):
                raise NotImplementedError("Parameters in the JDL are not supported")

            # TODO is this even needed?
            class_ad_job.insertAttributeInt("JobID", job_id)

            await check_and_prepare_job(
                job_id,
                class_ad_job,
                class_ad_req,
                job.owner,
                job.owner_group,
                job_attrs,
                job.vo,
                job_db,
                config,
            )
            job_jdl = createJDLWithInitialStatus(
                class_ad_job,
                class_ad_req,
                job_db.jdl_2_db_parameters,
                job_attrs,
                job.initial_status,
                job.initial_minor_status,
                modern=True,
            )

            jobs_to_insert[job_id] = job_attrs
            jdls_to_update[job_id] = compressJDL(job_jdl)

            if class_ad_job.lookupAttribute("InputData"):
                input_data = class_ad_job.getListFromExpression("InputData")
                inputdata_to_insert[job_id] = [lfn for lfn in input_data if lfn]

        tg.create_task(job_db.update_job_jdls(jdls_to_update))
        tg.create_task(job_db.insert_job_attributes(jobs_to_insert))

        if inputdata_to_insert:
            tg.create_task(job_db.insert_input_data(inputdata_to_insert))

    return list(jobs_to_insert.keys())
